-- TTCrx interface module
-- TTCrx register file occupies 128 bytes of memory space
-- to simplify the design, all accesses are 32 bit wide,
-- although only the LSB byte carries data.
-- according to TTCrx manual, its SCL line is input only.
-- So no arbitration and no wait holding are assumed in this design. 
library IEEE,mylib;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
use mylib.vtx_pack.all;
use mylib.dcc_pack.all;
library UNISIM;
use UNISIM.VComponents.all;
entity ttc_if is
    Port ( ttc_clk : in std_logic;
	 		  clk : in std_logic;
	 		  reset : in std_logic;
	 		  run : in std_logic;
	 		  n_strobe : in std_logic;
           rst_ttc : in std_logic;
           wr_ttc_id : in std_logic;
           rd_ttc_id : in std_logic;
           wr_i2c : in std_logic;
           rd_i2c : in std_logic;
           r_data : out std_logic_vector(31 downto 0);
           w_data : in std_logic_vector(31 downto 0);
           addr : in std_logic_vector(13 downto 0);
-- TTCrx status : ready, single error and double error
           en_brcst : in std_logic;
	 		  ttc_start : out std_logic;
	 		  ttc_stop : out std_logic;
	 		  ttc_soft_reset : out std_logic;
	 		  ttc_LRB_stat : out std_logic;
	 		  ttc_ready : out std_logic;
	 		  ttc_serr : out std_logic;
	 		  ttc_derr : out std_logic;
	 		  ttc_bcnt_err : out std_logic;
	 		  rate_OFW : out std_logic;
	 		  sync_lost : out std_logic;
           inc_oc : out std_logic;
           inc_l1ac : out std_logic;
           inc_bcnterr : out std_logic;
           inc_serr : out std_logic;
           inc_derr : out std_logic;
-- evn_fifo_full is controlled by strobe signal!!!
           state : in std_logic_vector(3 downto 0);
	 		  evn_fifo_full : in std_logic;
	 		  event_number_avl : out std_logic;
           event_number : out std_logic_vector(71 downto 0);
           ttc_pads : inout ttc_pad);
end ttc_if;

architecture ttc_arch of ttc_if is
signal ttc_brcststr1 : std_logic := '0';
signal ttc_brcst : std_logic_vector(3 downto 0) := (others => '0');
signal ttc_sinerrstr : std_logic := '0';
signal ttc_dberrstr : std_logic := '0';
signal ttc_l1accept : std_logic := '0';
signal ttc_bcntres : std_logic := '0';
signal ttc_evcntres : std_logic := '0';
signal ttc_OCres : std_logic := '0';
signal ttc_l1accept_dl : std_logic := '0';
signal ttc_bcntres_dl : std_logic := '0';
signal ttc_evcntres_dl : std_logic := '0';
signal ttc_OCres_dl : std_logic := '0';
signal dl_a : std_logic_vector(3 downto 0) := (others => '0');
signal dl_b : std_logic_vector(3 downto 0) := (others => '0');
signal bcnt : std_logic_vector(11 downto 0) := (others => '0');
signal evcnt : std_logic_vector(23 downto 0) := (others => '0');
signal oc : std_logic_vector(31 downto 0) := (others => '0');
signal strobe : std_logic := '0';
signal calib_evn : std_logic_vector(23 downto 0) := (others => '0');
signal fifo_din : std_logic_vector(71 downto 0) := (others => '0');
signal we_fifo : std_logic := '0';
--signal l1a_dl : std_logic := '0';
signal brcst_resetOC : std_logic := '0';
signal brcst_resync : std_logic := '0';
signal brcst_reset : std_logic := '0';
signal brcst_start : std_logic := '0';
signal brcst_statReq : std_logic := '0';
signal brcst_stop : std_logic := '0';
signal brcst_calibTrig : std_logic := '0';
signal reset_sync : std_logic_vector(2 downto 0) := (others => '0');
signal stop_sync : std_logic_vector(2 downto 0) := (others => '0');
signal start_sync : std_logic_vector(2 downto 0) := (others => '0');
signal l1a_toggle : std_logic := '0';
signal bcnterr_toggle : std_logic := '0';
signal serr_toggle : std_logic := '0';
signal derr_toggle : std_logic := '0';
signal LRB_stat_sync : std_logic_vector(2 downto 0) := (others => '0');
signal l1a_toggle_sync : std_logic_vector(2 downto 0) := (others => '0');
signal bcnterr_toggle_sync : std_logic_vector(2 downto 0) := (others => '0');
signal serr_toggle_sync : std_logic_vector(2 downto 0) := (others => '0');
signal derr_toggle_sync : std_logic_vector(2 downto 0) := (others => '0');
signal oc_toggle : std_logic := '0';
signal oc_sync : std_logic_vector(2 downto 0) := (others => '0');
signal serr : std_logic := '0';
signal derr : std_logic := '0';
signal clk_div : std_logic := '0';
signal ec_i2c : std_logic := '0';
signal i2c_busy : std_logic := '0';
signal ttc_id : std_logic_vector(31 downto 0) := (others => '0');
signal i2c_ad : std_logic_vector(15 downto 0) := (others => '0');
signal i2c_data : std_logic_vector(31 downto 0) := (others => '0');
signal i2c_active : std_logic := '0';
signal i2c_fail : std_logic := '0';
signal data_cycle : std_logic := '0';
signal i2c_phase : std_logic_vector(2 downto 0) := (others => '0');
signal i2c_done : std_logic := '0';
signal i2c_cntr : std_logic_vector(4 downto 0) := (others => '0');
signal rst_cnt : std_logic_vector(1 downto 0) := (others => '0');
signal scl : std_logic := '0';
signal sda : std_logic_vector(17 downto 0) := (others => '0');
signal eoc : std_logic := '0';
signal fifo_dout : std_logic_vector(71 downto 0) := (others => '0');
signal inc_fifo_wc : std_logic := '0';
signal rd_evn : std_logic := '0';
signal fifo_ovfl : std_logic := '0';
signal rate_OFWp : std_logic := '0';
signal dec_rate_cntr : std_logic := '0';
signal rate_div : std_logic_vector(8 downto 0) := (others => '0');
signal rate_cntr : std_logic_vector(5 downto 0) := (others => '0');
signal fifo_wa : std_logic_vector(8 downto 0) := (others => '0');
signal fifo_ra : std_logic_vector(8 downto 0) := (others => '0');
signal fifo_wc : std_logic_vector(3 downto 0) := (others => '0');
-- read back control
signal fifo_pntr : std_logic_vector(31 downto 0) := (others => '0');
signal event_data : std_logic_vector(31 downto 0) := (others => '0');
begin
inc_l1ac <= inc_fifo_wc;
event_number <= fifo_dout;
ttc_ready <= ttc_pads.ttcready;
ttc_serr <= serr;
ttc_derr <= derr;
ttc_pads.id <= ttc_id(13 downto 0);
ttc_id(31 downto 25) <= "0000000";
ttc_id(15 downto 14) <= "00";
process(clk)
begin
  if(clk'event and clk = '1')then
    if(strobe = '1')then
      if(wr_ttc_id = '1')then
        ttc_id(24 downto 16) <= w_data(24 downto 16);
        ttc_id(13 downto 0) <= w_data(13 downto 0);
      end if;
	 end if;
  end if;
end process;
-- reset ttcrx after ttc_id is valid
process(clk)
begin
  if(clk'event and clk = '1')then
    strobe <= n_strobe;
    if(strobe = '1' and rst_ttc = '1')then
	   ttc_pads.reset_b <= '0';
	 elsif(ec_i2c = '1' and rst_cnt(1) = '1')then
	   ttc_pads.reset_b <= '1';
	 end if;
    if(strobe = '1' and rst_ttc = '1')then
	   rst_cnt <= "00";
	 elsif(ec_i2c = '1')then
	   rst_cnt <= rst_cnt + 1;
	 end if;
  end if;
end process;
-- this divider scales down the clock to 1MHz
I_clk_div: l_cntr
  generic map(N=> 6)
  port map(
    d => gnd6,
    load => '0',
    reset => '0',
    set => '0',
    ec => strobe,
    en_cnt => '1',
    clk => clk,
    carry => clk_div
  );
process(clk)
begin
  if(clk'event and clk = '1')then
    ec_i2c <= clk_div and n_strobe;
  end if;
end process;
-- ttc broadcast commands
process(ttc_clk)
begin
  if(ttc_clk'event and ttc_clk = '1')then
    ttc_brcststr1 <=	ttc_pads.brcststr1;
	 ttc_brcst <= ttc_pads.brcst(7 downto 5) & ttc_pads.brcst(3);
    ttc_sinerrstr <= ttc_pads.sinerrstr;
    ttc_dberrstr <= ttc_pads.dberrstr;
	 ttc_l1accept <= ttc_pads.l1accept;
	 ttc_bcntres <= ttc_pads.bcntres;
	 ttc_evcntres <= ttc_pads.evcntres;
	 if(ttc_bcntres_dl = '1')then
	   bcnt <= x"000";
	 else
	   bcnt <= bcnt + 1;
	 end if;
	 if(ttc_bcntres_dl = '1' and bcnt /= x"deb")then
	   bcnterr_toggle <= not bcnterr_toggle;
	 end if;
	 if(ttc_evcntres_dl = '1')then
	   evcnt <= x"000001";
	 elsif(ttc_l1accept_dl = '1')then
	   evcnt <= evcnt + 1;
	 end if;
    if(ttc_brcststr1 = '1')then
--	   if(ttc_brcst = "0101")then
--        brcst_resync <= not brcst_resync;
--		end if;
	   if(ttc_brcst = "0111" or ttc_brcst = "0101")then
        brcst_reset <= not brcst_reset;
		end if;
	   if(ttc_brcst = "1001")then
        brcst_start <= not brcst_start;
		end if;
	   if(ttc_brcst = "1010")then
        brcst_statReq <= not brcst_statReq;
		end if;
	   if(ttc_brcst = "1011")then
        brcst_stop <= not brcst_stop;
		end if;
	 end if;
	 if(ttc_brcststr1 = '1' and ttc_brcst = "1101")then
      brcst_calibTrig <= '1';
	 else
      brcst_calibTrig <= '0';
	 end if;
	 if(ttc_OCres_dl = '1')then
	   oc <= x"0000000" & ttc_id(19 downto 16);
	 elsif(ttc_bcntres_dl = '1')then
	   oc <= oc + 1;
	 end if;
	 oc_toggle <= oc_toggle xor ttc_bcntres;
  end if;
end process;
-- adjusting Bcnt delay
dl_a <= ttc_id(23 downto 20) when ttc_id(24) = '1' else x"0";
dl_b <= ttc_id(23 downto 20) when ttc_id(24) = '0' else x"0";
ttc_OCres <= '1' when ttc_brcststr1 = '1' and ttc_brcst = "0011" else '0';
i_ttc_OCres_dl: srlc16 port map(
            d => ttc_OCres,
				clk => ttc_clk,
				a0 => dl_a(0),
				a1 => dl_a(1),
				a2 => dl_a(2),
				a3 => dl_a(3),
				q => ttc_OCres_dl);
i_ttc_bcntres_dl: srlc16 port map(
            d => ttc_bcntres,
				clk => ttc_clk,
				a0 => dl_a(0),
				a1 => dl_a(1),
				a2 => dl_a(2),
				a3 => dl_a(3),
				q => ttc_bcntres_dl);
i_ttc_l1accept_dl: srlc16 port map(
            d => ttc_l1accept,
				clk => ttc_clk,
				a0 => dl_b(0),
				a1 => dl_b(1),
				a2 => dl_b(2),
				a3 => dl_b(3),
				q => ttc_l1accept_dl);
i_ttc_evcntres_dl: srlc16 port map(
            d => ttc_evcntres,
				clk => ttc_clk,
				a0 => dl_b(0),
				a1 => dl_b(1),
				a2 => dl_b(2),
				a3 => dl_b(3),
				q => ttc_evcntres_dl);
process(clk)
begin
  if(clk'event and clk = '1')then
    if(strobe = '1')then
	   reset_sync <= reset_sync(1 downto 0) & brcst_reset;
	   stop_sync <= stop_sync(1 downto 0) & brcst_stop;
	   start_sync <= start_sync(1 downto 0) & brcst_start;
	   oc_sync <= oc_sync(1 downto 0) & oc_toggle;
		ttc_soft_reset <= (reset_sync(2) xor reset_sync(1)) and en_brcst;
		ttc_stop <= (stop_sync(2) xor stop_sync(1)) and en_brcst;
		ttc_start <= (start_sync(2) xor start_sync(1)) and en_brcst;
		inc_oc <= oc_sync(2) xor oc_sync(1);
	 end if;
  end if;
end process;
------------------------------------------------------------------------
process(ttc_clk)
begin
  if(ttc_clk'event and ttc_clk = '1')then
--	 fifo_din(35 downto 24) <= bcnt;
--	 fifo_din(71 downto 24) <= "000" & brcst_calibTrig & oc & bcnt;
	 fifo_din(71) <= state(3);
	 fifo_din(70) <= state(1) or (state(2) and not state(3));
	 fifo_din(69) <= state(2) xor state(0);
	 fifo_din(68 downto 24) <= brcst_calibTrig & oc & bcnt;
	 if(brcst_calibTrig = '1')then
	   fifo_din(23 downto 0) <= calib_evn;
	 else
	   fifo_din(23 downto 0) <= evcnt;
	 end if;
  end if;
end process;
--fifo_din(23 downto 0) <= evcnt when l1a_dl = '1' else calib_evn;
g_event_fifo: for i in 0 to 1 generate
  event_fifo: RAMB16_S36_S36
	  generic map(SIM_COLLISION_CHECK => "NONE")
    port map(DIA => fifo_din(i*36+31 downto i*36),
  			  DIPA => fifo_din(i*36+35 downto i*36+32),
  			  DIB => x"00000000",
  			  DIPB => x"0",
  			  ENA => '1',
  			  ENB => strobe,
  			  WEA => we_fifo,
  			  WEB => '0',
  			  SSRA => '0',
  			  SSRB => '0',
			  CLKA => ttc_clk,
			  CLKB => clk,
			  ADDRA => fifo_wa,
			  ADDRB => fifo_ra,
			  DOB => fifo_dout(i*36+31 downto i*36),
			  DOPB => fifo_dout(i*36+35 downto i*36+32)
			  );
end generate;
--fifo_din(71 downto 36) <= "000" & brcst_calibTrig & oc;
-- to accept a event number signals	l1accept,bcntstr,evcntlstr and evcntlstr
-- must all arrive at the right timing
process(ttc_clk,rst_ttc)
begin
  if(rst_ttc = '1')then
    l1a_toggle <= '0';
    serr_toggle <= '0';
    derr_toggle <= '0';
--	 l1a_dl <= '0';
  elsif(ttc_clk'event and ttc_clk = '1')then
-- lia_dl does not include calibration trigger, changed 2/16/2007
--    l1a_toggle <= l1a_toggle xor l1a_dl;
    l1a_toggle <= l1a_toggle xor we_fifo;
    serr_toggle <= serr_toggle xor ttc_sinerrstr;
    derr_toggle <= derr_toggle xor ttc_dberrstr;
--	 l1a_dl <= ttc_l1accept_dl;
  end if;
end process;
process(ttc_clk,reset)
begin
  if(reset = '1')then
	 serr <= '0';
	 derr <= '0';
	 calib_evn <= x"000001";
  elsif(ttc_clk'event and ttc_clk = '1')then
	 serr <= serr or ttc_sinerrstr;
	 derr <= derr or ttc_dberrstr;
	 if(brcst_calibTrig = '1')then
	   calib_evn <= calib_evn + 1;
	 end if;
  end if;
end process;
process(ttc_clk,reset)
begin
  if(reset = '1')then
    we_fifo <= '0';
    fifo_wa <= (others => '0');
    rate_div <= (others => '0');
    rate_cntr <= "000010";
	 rate_OFWp <= '0';
  elsif(ttc_clk'event and ttc_clk = '1')then
    we_fifo <= (ttc_l1accept_dl or brcst_calibTrig) and run;
    if(we_fifo = '1')then
      fifo_wa(3 downto 0) <= fifo_wa(3 downto 0) + 1;
	 end if;
	 if(rate_div = "110001110")then
	   dec_rate_cntr <= '1';
	 else
	   dec_rate_cntr <= '0';
	 end if;
	 if(dec_rate_cntr = '1')then
	   rate_div <= (others => '0');
	 else
	   rate_div <= rate_div + 1;
	 end if;
	 if(dec_rate_cntr /= we_fifo)then
	   if(we_fifo = '1')then
		  rate_cntr <= rate_cntr + 1;
		elsif(rate_cntr /= "000010")then
		  rate_cntr <= rate_cntr - 1;
		end if;
	 end if;
  	 if(rate_cntr(5) = '1')then
	   rate_OFWp <= '1';
	 elsif(rate_cntr(4) = '0')then
	   rate_OFWp <= '0';
	 end if;
end if;
end process;
process(clk)
begin
  if(clk'event and clk = '1')then
    if(reset = '1')then
	   fifo_ra <= (others => '0');
	   rate_OFW <= '0';
    elsif(strobe = '1')then
	   if(rd_evn = '1')then
		  fifo_ra(3 downto 0) <= fifo_ra(3 downto 0) + 1;
		end if;
	   rate_OFW <= rate_OFWp;
	 end if;
  end if;
end process;
rd_evn <= '0' when evn_fifo_full = '1' or fifo_wc = x"0" else '1'; 
I_fifo_wc: updown_cntr
    generic map (N => 4) 
    port map(
        rst => reset,
        set => '0',
        ce => strobe,
        up => inc_fifo_wc,
        down => rd_evn,
        clk => clk,
        q => fifo_wc,
		  carry => fifo_ovfl);
event_number_avl <= '0' when fifo_wc = x"0" else '1';
process(clk)
begin
  if(clk'event and clk = '1')then
    if(reset = '1')then
      l1a_toggle_sync <= "000";
		LRB_stat_sync <= "000";
      bcnterr_toggle_sync <= "000";
      serr_toggle_sync <= "000";
      derr_toggle_sync <= "000";
		inc_fifo_wc <= '0';
		inc_bcnterr <= '0';
		ttc_bcnt_err <= '0';
		inc_serr <= '0';
		inc_derr <= '0';
		sync_lost <= '0';
    elsif(strobe = '1')then
      l1a_toggle_sync <= l1a_toggle_sync(1 downto 0) & l1a_toggle;
		LRB_stat_sync <= LRB_stat_sync(1 downto 0) & brcst_statReq;
      bcnterr_toggle_sync <= bcnterr_toggle_sync(1 downto 0) & bcnterr_toggle;
      serr_toggle_sync <= serr_toggle_sync(1 downto 0) & serr_toggle;
      derr_toggle_sync <= derr_toggle_sync(1 downto 0) & derr_toggle;
		inc_fifo_wc <= l1a_toggle_sync(2) xor l1a_toggle_sync(1);
		if(bcnterr_toggle_sync(2) /= bcnterr_toggle_sync(1))then
		  inc_bcnterr <= '1';
		  ttc_bcnt_err <= '1';
		else
		  inc_bcnterr <= '0';
		end if;
		inc_serr <= serr_toggle_sync(2) xor serr_toggle_sync(1);
		inc_derr <= derr_toggle_sync(2) xor derr_toggle_sync(1);
		ttc_LRB_stat <= LRB_stat_sync(2) xor LRB_stat_sync(1);
		if(fifo_ovfl = '1' and evn_fifo_full = '1' and inc_fifo_wc = '1')then
		  sync_lost <= '1';
		end if;
	 end if;
  end if;
end process;
-- i2c access is initiated by a write to the even address of the i2c space
-- bits 31-16 are ignored. Bit 15 is '1' for read and '0' for write.
-- Bits 14-13 are ignored. Bits 12-8 are address of the the i2c register file.
-- Bits 7-0 are data to written and ignored for read operation.
process(clk)
begin
  if(clk'event and clk = '1')then
    if(reset = '1')then
	   i2c_busy <= '0';
    elsif(strobe = '1')then
	   i2c_busy <= wr_i2c or (i2c_busy and not i2c_done);
	 end if;
  end if;
end process;
process(clk)
begin
  if(clk'event and clk = '1')then
    if(reset = '1')then
	   i2c_done <= '0';
    elsif(ec_i2c = '1')then
		i2c_done <= eoc and (data_cycle or i2c_fail);
	 end if;
  end if;
end process;
process(clk)
begin
  if(clk'event and clk = '1')then
    if(strobe = '1')then
      if(wr_i2c = '1' and i2c_busy = '0')then
        i2c_ad <= w_data(15 downto 0);
      end if;
	 end if;
  end if;
end process;
-- i2c state machine
-- since this is the only master, no STOP is generated
process(clk)
begin
  if(clk'event and clk = '1')then
    if(reset = '1')then
		i2c_active <= '0';
		i2c_phase <= "000";
		data_cycle <= '0';
    elsif(ec_i2c = '1')then
	   i2c_active <= i2c_busy;
		i2c_phase(2 downto 1) <= i2c_phase(1 downto 0);
		i2c_phase(0) <= i2c_busy and (i2c_phase(2) or not i2c_active);
		data_cycle <= i2c_active and ((eoc and i2c_phase(2)) or data_cycle);
	 end if;
  end if;
end process;
-- scl generation and sda shift register 
process(clk)
begin
  if(clk'event and clk = '1')then
    if(ec_i2c = '1')then
      if(i2c_phase(1) = '1' and (i2c_cntr(4) = '0' or i2c_cntr(1) = '0'))then
	     i2c_data(7 downto 0) <= i2c_data(6 downto 0) & ttc_pads.sda;
	   end if;
	 end if;
  end if;
end process;
-- i2c_fail resets at the beginning of every i2c access
-- if acknoledge from slave is '1', i2c_fail is set and terminates the
-- operation
process(clk)
begin
  if(clk'event and clk = '1')then
    if(reset = '1')then
		i2c_fail <= '1';
    elsif(ec_i2c = '1')then
	   if(i2c_busy = '1' and i2c_active = '0')then
		  i2c_fail <= '0';
		elsif(ttc_pads.sda = '1' and i2c_phase(1) = '1')then
		  if(i2c_cntr = "01001" or
		    (i2c_cntr = "10010" and (data_cycle = '0' or i2c_ad(15) = '0')))then
		    i2c_fail <= '1';
		  end if;
		end if;   
	 end if;
  end if;
end process;
process(clk)
begin
  if(clk'event and clk = '1')then
    if(i2c_active = '0')then
	   scl <= '1';
		sda <= (others => '1');
	   eoc <= '0';
		i2c_cntr <= "00000";
    elsif(ec_i2c = '1')then
	   scl <= (not i2c_phase(1) and scl) or i2c_phase(0) or eoc;
	   if(i2c_phase(0) = '1' and i2c_cntr = "00000")then
		  sda(17 downto 8) <=  '0' & ttc_id(5 downto 0) & data_cycle & (i2c_ad(15) and data_cycle) & '1';
		  if(data_cycle = '0')then
		    sda(7 downto 0) <= "000" & i2c_ad(12 downto 8);
		  elsif(i2c_ad(15) = '1')then
		    sda(7 downto 0) <= "11111111"; 
		  else
		    sda(7 downto 0) <= i2c_ad(7 downto 0); 
		  end if;
		elsif(i2c_phase(2) = '1')then
		  sda <= sda(16 downto 0) & '1';
		end if;
	   if(i2c_phase(2) = '1')then
	     if(i2c_cntr = "10010" or i2c_fail = '1')then
		    eoc <= '1';
		  else
		    eoc <= '0';
		  end if;
		  if(eoc = '1')then
		    i2c_cntr <= "00000";
		  else
		    i2c_cntr <= i2c_cntr + 1;
		  end if;
		end if;
	 end if;
  end if;
end process;
-- sda and scl outputs
ttc_pads.sda <= '0' when sda(17) = '0' or eoc = '1' else 'Z';
ttc_pads.scl <= '0' when scl = '0' else 'Z';
-- read back i2c id or data
r_data <= ttc_id when rd_ttc_id = '1' else (others => 'Z');
-- i2c_data(13) is '1' if the last command was a write
-- i2c_data(7 downto 0) is valid read data only when all three MSBs are '0's
i2c_data(31 downto 8) <= x"0000" & i2c_busy & i2c_fail & not i2c_ad(15) & "00000";
r_data <= i2c_data when rd_i2c = '1' else (others => 'Z');
end ttc_arch;
